package com.flyfish.framework.approval.service;

import com.flyfish.framework.approval.domain.ApprovalDomain;
import com.flyfish.framework.approval.domain.record.ApproveRecord;
import com.flyfish.framework.approval.domain.submit.ApprovalSubmitDto;
import com.flyfish.framework.approval.domain.todo.ApprovalDto;
import com.flyfish.framework.approval.enums.ApproveAction;
import com.flyfish.framework.approval.enums.ApproveStatus;
import com.flyfish.framework.beans.meta.RestBean;
import com.flyfish.framework.exception.biz.InvalidBusinessException;
import com.flyfish.framework.service.BaseReactiveService;
import com.flyfish.framework.utils.ReflectionUtils;
import org.apache.commons.lang3.ClassUtils;
import org.springframework.beans.factory.ObjectProvider;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.annotation.AnnotationUtils;
import org.springframework.data.util.CastUtils;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import reactor.core.publisher.Mono;

import jakarta.annotation.Resource;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;

/**
 * rest模块代理服务
 * 代理所有带审批的bean服务，用于模糊审批
 *
 * @author wangyu
 */
@Service
public class ModuleDelegateService {

    @Resource
    private ApproveRecordService approveRecordService;

    // 审批的服务们
    private Map<String, BaseReactiveService<ApprovalDomain>> approvalServices;

    private String getModuleName(BaseReactiveService<ApprovalDomain> service) {
        return ReflectionUtils.getGenericType(service.getClass()).map(this::moduleName)
                .orElseThrow(() -> new InvalidBusinessException("系统模块名称抽取失败！"));
    }

    @Autowired
    @SuppressWarnings("all")
    public void setApprovalServices(ObjectProvider<BaseReactiveService> services) {
        this.approvalServices = services.stream()
                .filter(service -> ReflectionUtils.getGenericType(service.getClass())
                        .map(clazz -> ClassUtils.isAssignable(clazz, ApprovalDomain.class))
                        .orElse(false))
                .map(service -> (BaseReactiveService<ApprovalDomain>) service)
                .collect(Collectors.toMap(this::getModuleName, s -> s));
    }

    /**
     * 获取服务
     *
     * @param module 模块
     * @return 结果
     */
    public BaseReactiveService<ApprovalDomain> getService(String module) {
        return CastUtils.cast(approvalServices.get(module));
    }

    /**
     * 审批数据
     *
     * @param approval 记录
     * @return 结果
     */
    @Transactional(rollbackFor = Exception.class)
    public Mono<ApprovalDomain> approval(ApprovalDto approval) {
        return approveRecordService.getById(approval.getId())
                // 保存记录
                .flatMap(record -> {
                    if (record.getApproved()) {
                        return Mono.error(new InvalidBusinessException("该审批已被处理，请勿重复操作！"));
                    }
                    record.setApproved(true);
                    record.setAction(approval.getAction());
                    record.setOpinion(approval.getOpinion());
                    return approveRecordService.updateById(record);
                })
                // 修改主记录
                .flatMap(record -> {
                    BaseReactiveService<ApprovalDomain> service = getService(record.getModule());
                    return service.getById(record.getDataId())
                            .flatMap(domain -> {
                                // 如果已经审批过，直接返回
                                if (domain.getApproveStatus() == ApproveStatus.APPROVED) {
                                    return Mono.error(new InvalidBusinessException("该审批已被处理，请勿重复操作！"));
                                }
                                // 尚未提交审批，不处理
                                if (domain.getApproveStatus() == ApproveStatus.DRAFT || domain.getApproveStatus() == ApproveStatus.REJECTED) {
                                    return Mono.error(new InvalidBusinessException("该数据尚未提交审批，请确认后再操作！"));
                                }
                                // 终止标记
                                boolean end = false;
                                // 审批人
                                String approver = null;
                                // 尝试流转
                                switch (record.getAction()) {
                                    case AGREE:
                                        String next = determineNext(domain);
                                        // 下一个为空，代表审批完了，改状态
                                        if (null == next) {
                                            domain.setApproveStatus(ApproveStatus.APPROVED);
                                            domain.setNext(-1);
                                            end = true;
                                        } else {
                                            domain.setApproveStatus(ApproveStatus.APPROVING);
                                            domain.setNext(domain.getNext() + 1);
                                            approver = next;
                                        }
                                        break;
                                    case REJECT:
                                        // 拒绝，如果已经是第一步了，设置拒绝状态
                                        if (domain.getNext() == 0) {
                                            domain.setApproveStatus(ApproveStatus.REJECTED);
                                            end = true;
                                        } else {
                                            domain.setNext(domain.getNext() - 1);
                                            approver = domain.getApprovers().get(domain.getNext());
                                        }
                                        break;
                                    case REJECT_ALL:
                                        // 直接打回拒绝状态，需要重新修改提交
                                        domain.setNext(0);
                                        domain.setApproveStatus(ApproveStatus.REJECTED);
                                        end = true;
                                        break;
                                }
                                // 没结束
                                if (!end) {
                                    // 构建下一步操作
                                    ApproveRecord nextRecord = new ApproveRecord();
                                    nextRecord.setModule(record.getModule());
                                    nextRecord.setDataId(record.getId());
                                    nextRecord.setApproved(false);
                                    nextRecord.setModuleName(moduleName(domain.getClass()));
                                    nextRecord.setName(operation(approval.getAction()));
                                    nextRecord.setApprover(approver);
                                    return approveRecordService.create(nextRecord)
                                            .flatMap(r -> service.updateById(domain));
                                }
                                return service.updateById(domain);
                            });
                });

    }

    /**
     * 提交审核
     *
     * @param data 数据
     * @return 结果
     */
    public Mono<ApprovalDomain> submit(ApprovalSubmitDto data) {
        BaseReactiveService<ApprovalDomain> service = getService(data.getModule());
        return service.getById(data.getId())
                .flatMap(domain -> {
                    ApproveStatus status = domain.getApproveStatus();
                    if (null != status && ApproveStatus.DRAFT != status && ApproveStatus.REJECTED != status) {
                        return Mono.error(new InvalidBusinessException("该数据已经提交审批，请勿重复提交！"));
                    }
                    domain.setApprovers(data.getApprovers());
                    domain.setApproveStatus(ApproveStatus.PENDING);
                    domain.setNext(0);
                    return service.updateById(domain);
                })
                .flatMap(domain -> {
                    ApproveRecord record = new ApproveRecord();
                    record.setName("提交审核");
                    record.setModule(data.getModule());
                    record.setDataId(data.getId());
                    record.setApproved(false);
                    record.setApprover(domain.getApprovers().get(domain.getNext()));
                    record.setModuleName(moduleName(domain.getClass()));
                    return approveRecordService.create(record).thenReturn(domain);
                });
    }

    /**
     * 获取模块名称
     *
     * @param clazz 类
     * @return 结果
     */
    private String moduleName(Class<?> clazz) {
        return Optional.ofNullable(AnnotationUtils.findAnnotation(clazz, RestBean.class))
                .map(RestBean::name).orElse("未知模块");
    }

    /**
     * 决定下一位审批人
     *
     * @param domain 审批实体
     * @return 结果
     */
    private String determineNext(ApprovalDomain domain) {
        // 获取当前位置
        int index = domain.getNext();
        int next = index + 1;
        // 下一个位置
        if (next < domain.getApprovers().size()) {
            return domain.getApprovers().get(next);
        }
        return null;
    }

    /**
     * 操作中文名
     *
     * @param action 操作
     * @return 结果
     */
    private String operation(ApproveAction action) {
        return "将审批进行了\"" + action.getName() + "\"操作";
    }
}
