package com.flyfish.framework.backup.scheduler;

import com.flyfish.framework.backup.domain.Backup;
import com.flyfish.framework.context.DateContext;
import com.flyfish.framework.domain.base.DomainService;
import com.flyfish.framework.query.Query;
import com.flyfish.framework.repository.ReactiveEntityOperations;
import com.flyfish.framework.utils.JacksonUtil;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.ObjectProvider;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.core.io.buffer.DataBufferFactory;
import org.springframework.core.io.buffer.DataBufferUtils;
import org.springframework.core.io.buffer.DefaultDataBufferFactory;
import org.springframework.data.repository.core.EntityInformation;
import org.springframework.data.util.CastUtils;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;
import org.springframework.util.unit.DataSize;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

import java.io.IOException;
import java.nio.channels.AsynchronousFileChannel;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.time.LocalDateTime;
import java.util.List;
import java.util.UUID;
import java.util.concurrent.atomic.AtomicLong;
import java.util.stream.Collectors;

/**
 * 备份因功能小而简单，作为一个bean一直存在，定时跑就可以
 *
 * @author wangyu
 */
@Component
@Slf4j
public class BackupScheduler {

    // data buffer 工厂
    private final DataBufferFactory factory = new DefaultDataBufferFactory();
    // 用于注入所有集合名称
    private List<EntityInformation<?, String>> collections;
    // 异步的mongo操作，可以快速备份
    private ReactiveEntityOperations operations;
    // 备份路径
    private String backupPath = "/opt/flyfish/backup";

    @Autowired
    public void setServices(ObjectProvider<DomainService> services) {
        this.collections = services.stream().map(DomainService::getEntityInformation).collect(Collectors.toList());
    }

    @Autowired
    public void setOperations(ReactiveEntityOperations reactiveEntityOperations) {
        this.operations = reactiveEntityOperations;
    }

    @Autowired
    public void setBackupPath(@Value("${flyfish.backup.path:/opt/flyfish/backup}") String backupPath) {
        if (StringUtils.isNotBlank(backupPath)) {
            this.backupPath = backupPath;
        }
    }

    /**
     * 每天凌晨2点执行
     * 按日期建立目录，备份所有集合到对应目录下
     * 建立index.json，将绝对路径放入，遍历该文件即可还原
     */
    @Scheduled(cron = "0 0 2 * * ?")
    public void backup() {
        // 本次备份任务代号
        String code = UUID.randomUUID().toString();
        // 本次备份父目录，出错时替换目录，保证备份成功
        String parent = createIfNotExists(backupPath + "/" + code);
        // 开始备份，先构造一个指示器
        Backup backup = new Backup();
        backup.setCreateTime(LocalDateTime.now());
        backup.setCreator("系统");
        backup.setCode(code);
        backup.setStatus(Backup.Status.RUNNING);
        backup.setFilepath(parent + "/back.zip");
        // 数据大小计数器
        AtomicLong sizeCounter = new AtomicLong(0);
        // 开始备份
        operations.save(backup)
                .thenMany(Flux.fromIterable(this.collections))
                .flatMap(info -> operations
                        .findAll(Query.empty(), CastUtils.cast(info)).collectList()
                        .map(list -> new BackupIndex.BackupContent(info.getJavaType().getSimpleName(), JacksonUtil.toBytes(list)))
                )
                .flatMap(content -> writeContents(content, parent))
                .collectList()
                .flatMap(list -> {
                    BackupIndex index = new BackupIndex();
                    index.setId(code);
                    index.setItems(list.stream().map(content -> {
                        long currentSize = content.getContent().length;
                        sizeCounter.addAndGet(currentSize);
                        DataSize size = DataSize.ofBytes(currentSize);
                        BackupIndex.BackupItem item = new BackupIndex.BackupItem();
                        item.setCollection(content.getCollection());
                        item.setPath(parent + "/" + content.getCollection() + ".json");
                        item.setSize(size.toKilobytes() + "KB");
                        return item;
                    }).collect(Collectors.toList()));
                    // 写入备份结果
                    return write(JacksonUtil.toBytes(index), parent + "/meta.json");
                })
                .then(Mono.defer(() -> {
                    backup.setLog("成功备份");
                    backup.setPeriod(DateContext.distance(backup.getCreateTime(), LocalDateTime.now()));
                    backup.setStatus(Backup.Status.SUCCESS);
                    backup.setSize(DataSize.ofBytes(sizeCounter.get()).toKilobytes() + "KB");
                    return operations.save(backup);
                }))
                .subscribe(v -> {
                    log.info("成功完成备份");
                }, e -> {
                    backup.setStatus(Backup.Status.FAILED);
                    backup.setLog(e.getMessage());
                    backup.setPeriod(DateContext.distance(backup.getCreateTime(), LocalDateTime.now()));
                    operations.save(backup).subscribe();
                });

    }

    /**
     * 如果不存在，就创建
     *
     * @param filepath 路径
     */
    @SneakyThrows
    private String createIfNotExists(String filepath) {
        Path path = Paths.get(filepath);
        if (!Files.exists(path)) {
            try {
                Files.createDirectories(path);
            } catch (IOException e) {
                log.error(e.getMessage());
                return createIfNotExists(System.getProperty("user.home") + filepath);
            }
        }
        return filepath;
    }

    /**
     * 写入文件内容
     *
     * @param content 文件内容
     * @return 结果
     */
    private Mono<BackupIndex.BackupContent> writeContents(BackupIndex.BackupContent content, String parent) {
        String path = parent + "/" + content.getCollection() + ".json";
        return write(content.getContent(), path).thenReturn(content);
    }

    /**
     * 写入文件
     *
     * @param content 内容
     * @param path    路径
     * @return 结果
     */
    private Mono<Void> write(byte[] content, String path) {
        return Mono.fromCallable(() -> AsynchronousFileChannel.open(Files.createFile(Paths.get(path)), StandardOpenOption.WRITE))
                .flatMapMany(channel -> DataBufferUtils.write(Mono.just(factory.wrap(content)), channel, 0))
                .then();
    }

}
