package com.flyfish.framework.compiler.support;

import com.flyfish.framework.compiler.DynamicJavaCompiler;
import org.apache.commons.lang3.BooleanUtils;
import org.apache.commons.lang3.StringUtils;

import javax.tools.JavaCompiler;
import javax.tools.JavaFileManager;
import javax.tools.JavaFileObject;
import javax.tools.ToolProvider;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;

/**
 * 简单的java编译器，单文件，无缓存版
 * 只为了偷懒，快速撸代码
 *
 * @author wangyu
 */
public class SimpleJavaCompiler implements DynamicJavaCompiler {

    // 当前的java编译器
    private final JavaCompiler compiler = ToolProvider.getSystemJavaCompiler();
    // java文件管理器
    private final JavaFileManager fileManager = compiler.getStandardFileManager(null, Locale.SIMPLIFIED_CHINESE, StandardCharsets.UTF_8);
    // 编译参数
    private final List<String> options = Arrays.asList("-d", "./bin");
    // 编译池
    private final Map<String, Class<?>> classMap = new ConcurrentHashMap<>();

    private SimpleJavaCompiler() {
    }

    public static SimpleJavaCompiler getInstance() {
        return SingletonHolder.INSTANCE;
    }

    /**
     * 编译java源码，通过字符串
     *
     * @param name   文件名R
     * @param source 源码
     * @return 结果
     */
    @Override
    public String compile(String name, String source) {
        // 存在缓存，写入
        if (classMap.containsKey(name)) {
            return classMap.get(name).getName();
        }
        // 构建java文件对象
        JavaFileObject fileObject = new StringJavaObject(name, source);
        // 设置编译环境
        JavaCompiler.CompilationTask task = compiler.getTask(null, fileManager, null,
                options, null, Collections.singletonList(fileObject));
        // 阻塞编译
        if (BooleanUtils.isTrue(task.call())) {
            return "com.flyfish.project." + name;
        }
        return null;
    }

    /**
     * 加载类
     * 用于全部编译后，再加载
     *
     * @param name 名称
     * @return 结果
     */
    @Override
    public Class<?> load(String name) {
        String simpleName = StringUtils.substringAfterLast(name, ".");
        if (classMap.containsKey(simpleName)) {
            return classMap.get(simpleName);
        }
        try {
            Class<?> clazz = Class.forName(name);
            classMap.put(clazz.getSimpleName(), clazz);
            return clazz;
        } catch (ClassNotFoundException e) {
            return null;
        }
    }

    @Override
    public void close() throws IOException {

    }

    private static class SingletonHolder {

        private static final SimpleJavaCompiler INSTANCE = new SimpleJavaCompiler();

    }
}
