package com.flyfish.framework.compiler.template;

import com.flyfish.framework.compiler.support.JavaSource;
import org.springframework.beans.BeanUtils;
import org.springframework.data.util.CastUtils;
import org.springframework.util.StreamUtils;

import java.beans.PropertyDescriptor;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Method;
import java.nio.charset.StandardCharsets;
import java.util.Arrays;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * 模板编译器
 *
 * @author wangyu
 */
public class TemplateCompiler {

    // 变量池缓存
    private static final Map<Class<?>, Map<String, Method>> POOL_CACHE = new ConcurrentHashMap<>();

    /**
     * 编译模板文件和源码数据为源代码字符串
     *
     * @param dataClass 数据class
     * @param filename  文件名
     * @return 结果
     */
    public static <T> Function<T, String> compile(Class<T> dataClass, String filename) {
        // 获取文件内容
        InputStream fis = TemplateCompiler.class.getResourceAsStream(filename);
        try {
            // 获取模板内容
            String template = StreamUtils.copyToString(fis, StandardCharsets.UTF_8);
            // 准备变量池
            Map<String, Method> pool = methodPool(dataClass);
            // 编译
            return cfg -> pool.keySet().stream().reduce(template, (result, key) -> {
                // 获取读取方法
                Method method = pool.get(key);
                Object value = null;
                try {
                    // 取得结果
                    value = method.invoke(cfg);
                } catch (Exception ignored) {
                }
                // 替换变量
                return result.replace(key, parseValue(value));
            }, (a, b) -> a);
        } catch (IOException e) {
            e.printStackTrace();
            throw new RuntimeException("模板编译失败！");
        }
    }

    /**
     * 数据类方法池
     *
     * @param dataClass 数据类
     * @return 结果
     */
    private static Map<String, Method> methodPool(Class<?> dataClass) {
        return POOL_CACHE.computeIfAbsent(dataClass, k -> Arrays.stream(BeanUtils.getPropertyDescriptors(k))
                .collect(Collectors.toMap(descriptor -> "#{" + descriptor.getName() + "}", PropertyDescriptor::getReadMethod)));
    }

    /**
     * 解析单个值
     *
     * @param value 值
     * @return 结果
     */
    private static String parseValue(Object value) {
        if (null == value) {
            return "";
        }
        if (value instanceof Map) {
            Map<String, String> map = CastUtils.cast(value);
            return map.entrySet().stream()
                    .map(entry -> entry.getKey() + " " + entry.getValue())
                    .collect(Collectors.joining(" "));
        }
        return String.valueOf(value);
    }

    /**
     * 编译模板文件和源码数据为源代码字符串
     *
     * @param filename 文件名
     * @return 结果
     */
    public static Function<JavaSource, String> compile(String filename) {
        return compile(JavaSource.class, filename);
    }
}
