package com.flyfish.framework.query.chain;

import com.flyfish.framework.query.holder.QueryChainHolder;
import com.flyfish.framework.query.spi.adaptor.CriteriaAdaptor;
import com.flyfish.framework.utils.Assert;
import org.apache.commons.collections4.CollectionUtils;

import java.util.ArrayList;
import java.util.List;
import java.util.function.BiFunction;
import java.util.function.BinaryOperator;
import java.util.function.Function;

/**
 * 默认的查询保持器
 *
 * @param <C> 泛型，交给适配器处理
 */
class DefaultQueryChainHolder<C> implements QueryChainHolder<C> {

    private Function<CriteriaAdaptor<C>, BinaryOperator<C>> operator;

    private final List<BiFunction<C, CriteriaAdaptor<C>, C>> builders = new ArrayList<>();

    /**
     * 等待下一步消费
     * 如果不调用with，该状态将一直保持，可被其他link调用替换
     * 在调用with后，自动消费，并完成拼接
     *
     * @param operator 操作方法
     * @return 本身
     */
    @Override
    public QueryChainHolder<C> link(Function<CriteriaAdaptor<C>, BinaryOperator<C>> operator) {
        this.operator = operator;
        return this;
    }

    /**
     * 拼接条件
     *
     * @param criteria 条件
     */
    @Override
    public void with(Function<CriteriaAdaptor<C>, C> criteria) {
        Assert.notNull(operator, "连接操作不可为空");
        Function<CriteriaAdaptor<C>, BinaryOperator<C>> finalOperator = operator;
        this.operator = null;
        BiFunction<C, CriteriaAdaptor<C>, C> builder = (previous, adaptor) -> {
            // 连接逻辑
            BinaryOperator<C> linker = finalOperator.apply(adaptor);
            // 具体条件
            C next = criteria.apply(adaptor);
            // 之前的条件
            return linker.apply(previous, next);
        };
        builders.add(builder);
    }

    /**
     * 触发动作，得到最终的
     *
     * @return 本身
     */
    @Override
    public C get(CriteriaAdaptor<C> adaptor) {
        if (CollectionUtils.isNotEmpty(builders)) {
            return builders.stream()
                    .reduce(adaptor.empty(), (result, builder) -> builder.apply(result, adaptor), (a, b) -> a);
        }
        return adaptor.empty();
    }

    /**
     * 判断条件是否为空
     *
     * @return 结果
     */
    @Override
    public boolean isEmpty() {
        return builders.isEmpty();
    }
}
