package com.flyfish.framework.query.support;


import com.flyfish.framework.annotations.Property;
import com.flyfish.framework.utils.LRUCache;
import org.springframework.core.annotation.MergedAnnotations;
import org.springframework.util.ReflectionUtils;
import org.springframework.util.StringUtils;

import java.lang.ref.WeakReference;
import java.lang.reflect.Field;
import java.security.InvalidParameterException;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.ConcurrentHashMap;

public class EntityNameUtils {

    // SerializedLambda 反序列化缓存
    private static final Map<String, WeakReference<SerializedLambda>> FUNC_CACHE = new ConcurrentHashMap<>();

    // 列别名缓存
    private static final Map<Class<?>, Map<String, String>> COLUMN_CACHE = new LRUCache<>(5);

    private static final String MONGO_FIELD = "org.springframework.data.mongodb.core.mapping.Field";

    private static final String RELATIONAL_COLUMN = "org.springframework.data.relational.core.mapping.Column";

    /**
     * 方法转为属性
     *
     * @param name 方法名称
     * @return 最终属性
     */
    public static String methodToProperty(String name) {
        if (name.startsWith("is")) {
            name = name.substring(2);
        } else {
            if (!name.startsWith("get") && !name.startsWith("set")) {
                throw new InvalidParameterException("不是正确的getter或setter");
            }
            name = name.substring(3);
        }
        if (name.length() == 1 || name.length() > 1 && !Character.isUpperCase(name.charAt(1))) {
            name = name.substring(0, 1).toLowerCase(Locale.ENGLISH) + name.substring(1);
        }
        return name;
    }

    /**
     * 将驼峰命名转化成下划线
     *
     * @param param 参数
     * @return 结果
     */
    public static String camelToUnderline(String param) {
        if (param.length() < 3) {
            return param.toLowerCase();
        }
        StringBuilder sb = new StringBuilder(param);
        int temp = 0;//定位
        //从第三个字符开始 避免命名不规范
        for (int i = 2; i < param.length(); i++) {
            if (Character.isUpperCase(param.charAt(i))) {
                sb.insert(i + temp, "_");
                temp += 1;
            }
        }
        return sb.toString().toLowerCase();
    }

    /**
     * 解析列并使用处理函数处理
     *
     * @param func 方法引用
     * @param <T>  实体泛型
     * @return 处理结果
     */
    public static <T> String getName(SFunction<T, ?> func) {
        SerializedLambda lambda = resolve(func);
        String property = methodToProperty(lambda.getImplMethodName());
        Class<?> beanClass = resolveEntityClass(lambda);
        return tryCache(beanClass).getOrDefault(property, property);
    }

    /**
     * 解析获得实体类
     *
     * @param lambda 序列化的lambda
     * @return 最终获取的类
     */
    private static Class<?> resolveEntityClass(SerializedLambda lambda) {
        Class<?> type = lambda.getInstantiatedType();
        tryCache(type);
        return type;
    }

    /**
     * 尝试缓存
     *
     * @param entityClass bean的类型
     */
    private static Map<String, String> tryCache(Class<?> entityClass) {
        return COLUMN_CACHE.computeIfAbsent(entityClass, EntityNameUtils::buildFieldsCache);
    }


    /**
     * 构建字段缓存
     *
     * @param type 类型
     * @return 构建后的缓存
     */
    private static Map<String, String> buildFieldsCache(Class<?> type) {
        Map<String, String> fields = new HashMap<>();
        ReflectionUtils.doWithFields(type, field -> fields.put(field.getName(), resolveFinalName(field)));
        return fields;
    }


    /**
     * 解析字段注解或直接取用下划线逻辑
     *
     * @return 解析结果
     */
    private static String resolveFinalName(Field field) {
        MergedAnnotations annotations = MergedAnnotations.from(field);
        if (annotations.isPresent(Property.class)) {
            String key = annotations.get(Property.class).getString("key");
            if (StringUtils.hasText(key)) {
                return key;
            }
        }
        if (annotations.isPresent(MONGO_FIELD)) {
            String name = annotations.get(MONGO_FIELD).getString("name");
            if (StringUtils.hasText(name)) {
                return name;
            }
        }
        if (annotations.isPresent(RELATIONAL_COLUMN)) {
            String name = annotations.get(RELATIONAL_COLUMN).getString("value");
            if (StringUtils.hasText(name)) {
                return name;
            }
        }
        return field.getName();
    }

    /**
     * 解析方法引用为序列化lambda实例
     * 该方式会使用缓存
     *
     * @param func 方法引用
     * @param <T>  泛型
     * @return 解析结果
     */
    private static <T> SerializedLambda resolve(SFunction<T, ?> func) {
        Class<?> clazz = func.getClass();
        String name = clazz.getName();
        return Optional.ofNullable(FUNC_CACHE.get(name))
                .map(WeakReference::get)
                .orElseGet(() -> {
                    SerializedLambda lambda = SerializedLambda.resolve(func);
                    FUNC_CACHE.put(name, new WeakReference<>(lambda));
                    return lambda;
                });
    }
}
