package com.flyfish.framework.r2dbc.config;

import com.flyfish.framework.domain.base.Domain;
import com.flyfish.framework.r2dbc.config.callback.ReferenceR2dbcCallback;
import com.flyfish.framework.r2dbc.config.convert.AdditionalConverters;
import com.flyfish.framework.r2dbc.config.convert.FullMappingR2dbcConverter;
import com.flyfish.framework.r2dbc.config.convert.FullReactiveDataAccessStrategy;
import com.flyfish.framework.r2dbc.metadata.R2dbcMetadataManager;
import com.flyfish.framework.r2dbc.metadata.impl.SimpleR2dbcMetadataManager;
import com.flyfish.framework.r2dbc.operations.R2dbcReactiveEntityOperations;
import com.flyfish.framework.r2dbc.repository.factory.DefaultReactiveRepositoryFactoryBean;
import com.flyfish.framework.r2dbc.repository.impl.QueryingR2dbcReactiveRepository;
import com.flyfish.framework.repository.ReactiveEntityOperations;
import io.r2dbc.spi.ConnectionFactory;
import org.apache.commons.lang3.ArrayUtils;
import org.springframework.context.annotation.Bean;
import org.springframework.core.io.Resource;
import org.springframework.core.io.support.PathMatchingResourcePatternResolver;
import org.springframework.core.io.support.ResourcePatternResolver;
import org.springframework.data.convert.CustomConversions;
import org.springframework.data.mapping.callback.EntityCallback;
import org.springframework.data.r2dbc.config.EnableR2dbcAuditing;
import org.springframework.data.r2dbc.convert.MappingR2dbcConverter;
import org.springframework.data.r2dbc.convert.R2dbcConverter;
import org.springframework.data.r2dbc.convert.R2dbcCustomConversions;
import org.springframework.data.r2dbc.core.R2dbcEntityOperations;
import org.springframework.data.r2dbc.core.R2dbcEntityTemplate;
import org.springframework.data.r2dbc.core.ReactiveDataAccessStrategy;
import org.springframework.data.r2dbc.dialect.DialectResolver;
import org.springframework.data.r2dbc.dialect.R2dbcDialect;
import org.springframework.data.r2dbc.mapping.R2dbcMappingContext;
import org.springframework.data.r2dbc.repository.config.EnableR2dbcRepositories;
import org.springframework.r2dbc.connection.init.CompositeDatabasePopulator;
import org.springframework.r2dbc.connection.init.ConnectionFactoryInitializer;
import org.springframework.r2dbc.connection.init.DatabasePopulator;
import org.springframework.r2dbc.connection.init.ResourceDatabasePopulator;
import org.springframework.r2dbc.core.DatabaseClient;
import org.springframework.util.Assert;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Optional;

/**
 * r2dbc数据配置
 *
 * @author wangyu
 */
@EnableR2dbcRepositories(
        repositoryFactoryBeanClass = DefaultReactiveRepositoryFactoryBean.class,
        repositoryBaseClass = QueryingR2dbcReactiveRepository.class,
        basePackages = "com.flyfish.framework.*.repository"
)
@EnableR2dbcAuditing
public class R2dbcDataConfig {

    private final DatabaseClient databaseClient;
    private final R2dbcDialect dialect;

    public R2dbcDataConfig(DatabaseClient databaseClient) {
        this.databaseClient = databaseClient;
        this.dialect = DialectResolver.getDialect(databaseClient.getConnectionFactory());
    }

    @Bean
    @SuppressWarnings("deprecation")
    public R2dbcEntityTemplate r2dbcEntityTemplate(ReactiveDataAccessStrategy reactiveDataAccessStrategy) {
        return new R2dbcEntityTemplate(this.databaseClient, reactiveDataAccessStrategy);
    }

    @Bean
    public ReactiveEntityOperations r2dbcReactiveEntityOperations(R2dbcEntityOperations r2dbcEntityOperations) {
        return new R2dbcReactiveEntityOperations(r2dbcEntityOperations);
    }

    @Bean
    public EntityCallback<Domain> referenceR2dbcCallback(R2dbcMetadataManager metadataManager) {
        return new ReferenceR2dbcCallback(metadataManager);
    }

    @Bean
    @SuppressWarnings("deprecation")
    public ReactiveDataAccessStrategy reactiveDataAccessStrategy(R2dbcConverter converter) {
        Assert.notNull(converter, "MappingContext must not be null");
        return new FullReactiveDataAccessStrategy(dialect, converter);
    }

    @Bean
    public MappingR2dbcConverter r2dbcConverter(R2dbcMappingContext mappingContext, R2dbcCustomConversions r2dbcCustomConversions) {
        return new FullMappingR2dbcConverter(mappingContext, r2dbcCustomConversions);
    }

    @Bean
    public R2dbcMetadataManager r2dbcMetadataManager(R2dbcMappingContext mappingContext) {
        return new SimpleR2dbcMetadataManager(mappingContext);
    }

    @Bean
    public R2dbcCustomConversions additionalConversions() {
        List<Object> converters = new ArrayList(this.dialect.getConverters());
        converters.addAll(R2dbcCustomConversions.STORE_CONVERTERS);
        converters.addAll(AdditionalConverters.ADDING_CONVERTERS);
        return new R2dbcCustomConversions(CustomConversions.StoreConversions.of(this.dialect.getSimpleTypeHolder(), converters),
                Collections.emptyList());
    }

    /**
     * 数据库初始化
     *
     * @param connectionFactory 链接仓库
     * @return 结果
     */
    @Bean
    public ConnectionFactoryInitializer connectionFactoryInitializer(ConnectionFactory connectionFactory) {
        ConnectionFactoryInitializer initializer = new ConnectionFactoryInitializer();
        initializer.setConnectionFactory(connectionFactory);
        CompositeDatabasePopulator populator = new CompositeDatabasePopulator();
        // 添加主要脚本
        preparePopulator("classpath*:schema.sql").ifPresent(populator::addPopulators);
        // 动态执行方言脚本
        preparePopulator("classpath*:dialect/" + connectionFactory.getMetadata().getName().toLowerCase() + ".sql")
                .ifPresent(populator::addPopulators);
        initializer.setDatabasePopulator(populator);
        // 添加清理脚本
        preparePopulator("classpath*:schema-clean.sql").ifPresent(initializer::setDatabaseCleaner);
        // 返回初始化器
        return initializer;
    }

    /**
     * 动态判定存在性，提供健壮的架构
     *
     * @param path 文件路径
     * @return 结果
     */
    private Optional<DatabasePopulator> preparePopulator(String path) {
        ResourcePatternResolver resolver = new PathMatchingResourcePatternResolver();
        try {
            Resource[] resources = resolver.getResources(path);
            if (ArrayUtils.isNotEmpty(resources)) {
                return Optional.of(new ResourceDatabasePopulator(resources));
            }
        } catch (IOException ignored) {
        }
        return Optional.empty();
    }
}
