package com.flyfish.framework.r2dbc.config.callback;

import com.flyfish.framework.domain.base.Domain;
import com.flyfish.framework.query.Queries;
import com.flyfish.framework.r2dbc.metadata.reference.R2dbcAssociation;
import com.flyfish.framework.r2dbc.metadata.reference.R2dbcCollection;
import com.flyfish.framework.r2dbc.metadata.reference.R2dbcRelation;
import com.flyfish.framework.r2dbc.metadata.reference.field.FieldAccessor;
import com.flyfish.framework.r2dbc.metadata.visitor.MetadataHandler;
import com.flyfish.framework.repository.DefaultReactiveRepository;
import lombok.RequiredArgsConstructor;
import org.apache.commons.lang3.ObjectUtils;
import org.springframework.data.mapping.PersistentProperty;
import org.springframework.data.mapping.PersistentPropertyAccessor;
import org.springframework.data.r2dbc.core.R2dbcEntityOperations;
import org.springframework.data.r2dbc.core.StatementMapper;
import org.springframework.data.relational.core.mapping.RelationalPersistentProperty;
import org.springframework.data.relational.core.query.Criteria;
import org.springframework.r2dbc.core.DatabaseClient;
import org.springframework.r2dbc.core.PreparedOperation;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

import java.util.function.Supplier;

@RequiredArgsConstructor
class AssociationFetcher implements MetadataHandler {

    private final RepositoryProvider repositories;

    private final Supplier<R2dbcEntityOperations> operations;

    /**
     * 处理association
     *
     * @param association 一对一关联
     * @param accessor    对象访问器
     * @return 结果
     */
    @Override
    public Mono<Domain> handleAssociation(R2dbcAssociation association, PersistentPropertyAccessor<Domain> accessor) {
        // 准备查询， 获取对方的仓库
        DefaultReactiveRepository<Domain> repository = repositories.get(association.getEntity());
        // 准备设置器
        FieldAccessor field = association.getAccessor();
        if (association.isInner()) {
            // 处理内部查询，以对象内的值作为查询条件
            Object value = accessor.getProperty(association.getField());
            if (ObjectUtils.isEmpty(value)) return null;
            // 查询
            return repository.findById(String.valueOf(value))
                    .map(result -> field.set(accessor.getBean(), result))
                    .defaultIfEmpty(accessor.getBean());
        } else {
            // 处理外部关联查询
            String id = accessor.getBean().getId();
            // 直接通过外部字段查询
            return repository.findOne(Queries.where(getColumnName(association.getField())).eq(id).wrap())
                    .map(result -> field.set(accessor.getBean(), result))
                    .defaultIfEmpty(accessor.getBean());
        }
    }

    /**
     * 处理collection
     *
     * @param collection 一对多关联
     * @param accessor   对象访问器
     * @return 结果
     */
    @Override
    public Mono<Domain> handleCollection(R2dbcCollection collection, PersistentPropertyAccessor<Domain> accessor) {
        // 获取条件
        String id = accessor.getBean().getId();
        // 准备查询
        DefaultReactiveRepository<Domain> repository = repositories.get(collection.getEntity());
        // 准备设置
        FieldAccessor field = collection.getAccessor();
        // 直接通过外部字段查询
        return repository.findAll(Queries.where(getColumnName(collection.getField())).eq(id).wrap())
                .collectList()
                .map(result -> field.set(accessor.getBean(), result))
                .defaultIfEmpty(accessor.getBean());
    }

    /**
     * 处理relation
     *
     * @param relation 多对多关联
     * @param accessor 访问器
     * @return 结果
     */
    @Override
    public Mono<Domain> handleRelation(R2dbcRelation relation, PersistentPropertyAccessor<Domain> accessor) {
        // 准备目标对象的repo
        DefaultReactiveRepository<Domain> repository = repositories.get(relation.getEntity());
        // 准备查询参数
        String id = accessor.getBean().getId();
        // 查询关联id集合
        Flux<String> ids = fetchRelationIds(relation, id);
        // 准备设置
        FieldAccessor field = relation.getAccessor();
        // 查询最终结果
        return repository.findAllById(ids)
                .collectList()
                .map(result -> field.set(accessor.getBean(), result))
                .defaultIfEmpty(accessor.getBean());
    }


    /**
     * 查询当前实体关联的ids
     *
     * @param relation 关联信息
     * @return 结果
     */
    @SuppressWarnings("deprecation")
    private Flux<String> fetchRelationIds(R2dbcRelation relation, String id) {
        // 准备查询器
        R2dbcEntityOperations entityOperations = operations.get();
        StatementMapper statementMapper = entityOperations.getDataAccessStrategy().getStatementMapper();
        DatabaseClient databaseClient = entityOperations.getDatabaseClient();
        // 构建查询
        StatementMapper.SelectSpec select = statementMapper.createSelect(relation.getTableName())
                .withProjection(relation.getForeignField()).withCriteria(Criteria.where(relation.getField()).is(id));
        PreparedOperation<?> operation = statementMapper.getMappedObject(select);
        // 执行查询
        return databaseClient.sql(operation)
                .map(row -> row.get(0, String.class))
                .all();
    }

    /**
     * 通过持久化属性获取真实的列名
     *
     * @param property 属性
     * @return 结果
     */
    private String getColumnName(PersistentProperty<?> property) {
        if (property instanceof RelationalPersistentProperty) {
            return ((RelationalPersistentProperty) property).getColumnName().getReference();
        }
        return property.getName();
    }
}
