package com.flyfish.framework.r2dbc.config.convert;

import com.flyfish.framework.utils.JacksonUtil;
import org.springframework.data.r2dbc.convert.R2dbcConverter;
import org.springframework.data.r2dbc.core.DefaultReactiveDataAccessStrategy;
import org.springframework.data.r2dbc.dialect.R2dbcDialect;
import org.springframework.data.r2dbc.mapping.OutboundRow;
import org.springframework.data.relational.core.mapping.RelationalPersistentEntity;
import org.springframework.data.relational.core.mapping.RelationalPersistentProperty;
import org.springframework.lang.NonNull;
import org.springframework.r2dbc.core.Parameter;
import org.springframework.util.Assert;
import org.springframework.util.ClassUtils;

import java.util.Collection;

/**
 * 全支持的数据访问策略，将数组扁平化转换，方便支持json
 *
 * @author wangyu
 */
@SuppressWarnings("deprecation")
public class FullReactiveDataAccessStrategy extends DefaultReactiveDataAccessStrategy {

    private final R2dbcConverter converter;

    private final R2dbcDialect dialect;

    /**
     * 最大化配置的构造器，仅提供该实现
     *
     * @param dialect   别名实现
     * @param converter 转换器
     */
    public FullReactiveDataAccessStrategy(R2dbcDialect dialect, R2dbcConverter converter) {
        super(dialect, converter);
        this.converter = converter;
        this.dialect = dialect;
    }

    @Override
    @NonNull
    public OutboundRow getOutboundRow(@NonNull Object object) {
        Assert.notNull(object, "Entity object must not be null");

        OutboundRow row = new OutboundRow();

        converter.write(object, row);

        RelationalPersistentEntity<?> entity = getRequiredPersistentEntity(ClassUtils.getUserClass(object));

        for (RelationalPersistentProperty property : entity) {
            Parameter value = row.get(property.getColumnName());
            if (value != null && shouldConvertArrayValue(property, value)) {
                Parameter writeValue = getArrayValue(value);
                row.put(property.getColumnName(), writeValue);
            }
        }

        return row;
    }

    private Parameter getArrayValue(Parameter value) {
        if (value.getType().equals(byte[].class)) {
            return value;
        }
        return Parameter.fromOrEmpty(JacksonUtil.toJson(value.getValue()).orElse(null), String.class);
    }

    /**
     * 判断是否需要转换数据值
     *
     * @param property 属性
     * @param value    值表示
     * @return 结果
     */
    private boolean shouldConvertArrayValue(RelationalPersistentProperty property, Parameter value) {

        if (!property.isCollectionLike()) {
            return false;
        }

        if (value.hasValue() && (value.getValue() instanceof Collection || value.getValue().getClass().isArray())) {
            return true;
        }

        if (Collection.class.isAssignableFrom(value.getType()) || value.getType().isArray()) {
            return true;
        }

        return false;
    }

    @SuppressWarnings("unchecked")
    private RelationalPersistentEntity<?> getRequiredPersistentEntity(Class<?> typeToRead) {
        return converter.getMappingContext().getRequiredPersistentEntity(typeToRead);
    }

}
