package com.flyfish.framework.r2dbc.metadata.visitor;

import com.flyfish.framework.domain.base.Domain;
import com.flyfish.framework.r2dbc.metadata.R2dbcTableMetadata;
import com.flyfish.framework.r2dbc.metadata.reference.R2dbcAssociation;
import com.flyfish.framework.r2dbc.metadata.reference.R2dbcCollection;
import com.flyfish.framework.r2dbc.metadata.reference.R2dbcRelation;
import lombok.Getter;
import org.apache.commons.collections4.CollectionUtils;
import org.springframework.data.mapping.PersistentEntity;
import org.springframework.data.mapping.PersistentPropertyAccessor;
import reactor.core.publisher.Mono;

import java.util.List;
import java.util.Objects;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * r2dbc表元数据访问器
 *
 * @author wangyu
 */
public final class R2dbcTableMetadataVisitor implements MetadataVisitor {

    @Getter
    private Mono<Domain> result;

    private final Domain entity;

    private final MetadataHandler handler;

    private PersistentPropertyAccessor<Domain> accessor;

    private R2dbcTableMetadata metadata;

    public R2dbcTableMetadataVisitor(Domain entity, MetadataHandler handler) {
        this.entity = entity;
        this.handler = handler;
        this.result = Mono.just(entity);
    }

    /**
     * 访问元数据的方法
     *
     * @param metadata 元数据
     */
    @Override
    public void enter(R2dbcTableMetadata metadata) {
        // 不含有关联关系，直接跳过
        if (!metadata.hasAssociation()) {
            return;
        }
        this.metadata = metadata;
        // 获取本实体的持久化对象
        PersistentEntity<?, ?> persistentEntity = metadata.getEntity();
        // 获取属性设置器
        this.accessor = persistentEntity.getPropertyAccessor(entity);

        // 尝试填充各种关联方式
        List<Mono<Domain>> signals = Stream.of(visitAssociations(), visitCollections(), visitRelations())
                .flatMap(Function.identity())
                .filter(Objects::nonNull)
                .collect(Collectors.toList());

        // 尝试填充
        if (CollectionUtils.isNotEmpty(signals)) {
            this.result = Mono.zip(signals, objs -> entity);
        }
    }


    /**
     * 访问一对一关联
     *
     * @return 结果
     */
    private Stream<Mono<Domain>> visitAssociations() {
        List<R2dbcAssociation> associations = metadata.getAssociations();
        if (CollectionUtils.isNotEmpty(associations)) {
            return associations.stream().map(association -> handler.handleAssociation(association, accessor));
        }
        return Stream.empty();
    }

    /**
     * 拉取一对多关联实体
     *
     * @return 结果
     */
    private Stream<Mono<Domain>> visitCollections() {
        List<R2dbcCollection> collections = metadata.getCollections();
        if (CollectionUtils.isNotEmpty(collections)) {
            return collections.stream().map(collection -> handler.handleCollection(collection, accessor));
        }
        return Stream.empty();
    }

    /**
     * 拉取多对多关联数据
     *
     * @return 结果
     */
    private Stream<Mono<Domain>> visitRelations() {
        List<R2dbcRelation> relations = metadata.getRelations();
        if (CollectionUtils.isNotEmpty(relations)) {
            return relations.stream().map(relation -> handler.handleRelation(relation, accessor));
        }
        return Stream.empty();
    }
}
