package com.flyfish.framework.dict.handler;

import com.flyfish.framework.annotations.DictValue;
import com.flyfish.framework.context.interceptor.ViewModelInterceptor;
import com.flyfish.framework.dict.domain.DictionaryValue;
import com.flyfish.framework.dict.service.DictionaryService;
import com.flyfish.framework.utils.ReflectionUtils;
import lombok.RequiredArgsConstructor;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.reflect.FieldUtils;
import org.springframework.stereotype.Component;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;
import reactor.util.function.Tuples;

import java.lang.reflect.Field;
import java.util.List;
import java.util.Objects;

import static org.apache.commons.lang3.ObjectUtils.defaultIfNull;

/**
 * 保证首当其冲被检测到，设置order 0
 * 用于处理字典值，无侵入注入
 *
 * @author wangyu
 */
@Component
@RequiredArgsConstructor
public class DictAnnotationInterceptor implements ViewModelInterceptor {

    // 字典服务
    private final DictionaryService dictionaryService;

    /**
     * 当前对象包含@DictValue的属性时支持
     *
     * @param data 返回值
     * @return 结果
     */
    @Override
    public boolean supports(Object data) {
        List<Field> fields = FieldUtils.getFieldsListWithAnnotation(data.getClass(), DictValue.class);
        return CollectionUtils.isNotEmpty(fields);
    }

    /**
     * 处理对象中的字典注解
     *
     * @param data 数据
     * @return 处理后结果
     */
    @Override
    public <T> Mono<T> intercept(T data) {
        List<Field> fields = FieldUtils.getFieldsListWithAnnotation(data.getClass(), DictValue.class);
        return Flux.fromIterable(fields)
                .flatMap(field -> {
                    DictValue dictValue = field.getAnnotation(DictValue.class);
                    String code = dictValue.value();
                    String fieldName = field.getName();
                    Object value = ReflectionUtils.getFieldValue(data, fieldName);
                    return dictionaryService.getByCode(code).map(dict -> dict.getValues()
                            .stream()
                            .filter(item -> item.getValue().equals(value))
                            .findFirst()
                            .map(DictionaryValue::getText)
                            .orElse(defaultIfNull((String) value, ""))
                    ).map(mapped -> Tuples.of(fieldName, mapped));
                })
                .filter(Objects::nonNull)
                .reduce(data, (result, tuple) -> {
                    ReflectionUtils.setFieldValue(result, tuple.getT1(), tuple.getT2());
                    return result;
                });
    }
}
