package com.flyfish.framework.configuration;

import com.fasterxml.jackson.annotation.JsonInclude;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.flyfish.framework.configuration.resolver.PageQueryArgumentResolver;
import com.flyfish.framework.configuration.resolver.RequestContextBodyArgumentResolver;
import com.flyfish.framework.configuration.resolver.UserArgumentResolver;
import com.flyfish.framework.configuration.resolver.ValidRequestBodyMethodArgumentResolver;
import com.flyfish.framework.context.SpringContext;
import jakarta.annotation.Resource;
import org.springframework.boot.autoconfigure.jackson.JacksonProperties;
import org.springframework.context.annotation.Configuration;
import org.springframework.core.Ordered;
import org.springframework.core.ReactiveAdapterRegistry;
import org.springframework.core.annotation.Order;
import org.springframework.http.codec.CodecConfigurer;
import org.springframework.http.codec.ServerCodecConfigurer;
import org.springframework.http.codec.json.Jackson2JsonDecoder;
import org.springframework.http.codec.json.Jackson2JsonEncoder;
import org.springframework.http.converter.json.Jackson2ObjectMapperBuilder;
import org.springframework.web.reactive.config.EnableWebFlux;
import org.springframework.web.reactive.config.WebFluxConfigurer;
import org.springframework.web.reactive.result.method.annotation.ArgumentResolverConfigurer;

import java.util.Optional;
import java.util.TimeZone;

/**
 * token拦截器配置
 *
 * @author Mr.Wang
 */
@Configuration
@EnableWebFlux
@Order(Ordered.HIGHEST_PRECEDENCE)
public class WebfluxConfig implements WebFluxConfigurer {

    @Resource
    private JacksonProperties jacksonProperties;

    /**
     * Configure resolvers for custom {@code @RequestMapping} method arguments.
     *
     * @param configurer to configurer to use
     */
    @Override
    public void configureArgumentResolvers(ArgumentResolverConfigurer configurer) {
        ReactiveAdapterRegistry registry = ReactiveAdapterRegistry.getSharedInstance();
        ServerCodecConfigurer serverCodecConfigurer = SpringContext.getBean(ServerCodecConfigurer.class);
        configurer.addCustomResolver(new PageQueryArgumentResolver(registry));
        configurer.addCustomResolver(new UserArgumentResolver(registry));
        configurer.addCustomResolver(new ValidRequestBodyMethodArgumentResolver(serverCodecConfigurer.getReaders(), registry));
        configurer.addCustomResolver(new RequestContextBodyArgumentResolver(serverCodecConfigurer.getReaders(), registry));
    }

    /**
     * 配置jackson
     *
     * @param configurer 配置器
     */
    @Override
    public void configureHttpMessageCodecs(ServerCodecConfigurer configurer) {
        // 设置时区
        TimeZone.setDefault(TimeZone.getTimeZone("GMT+8"));
        CodecConfigurer.DefaultCodecs defaults = configurer.defaultCodecs();
        ObjectMapper mapper = Jackson2ObjectMapperBuilder.json()
                .simpleDateFormat(Optional.ofNullable(jacksonProperties.getDateFormat()).orElse("yyyy-MM-dd HH:mm:ss"))
                .timeZone("GMT+8")
                .serializationInclusion(Optional.ofNullable(jacksonProperties.getDefaultPropertyInclusion())
                        .orElse(JsonInclude.Include.NON_NULL))
                .build();
        defaults.jackson2JsonDecoder(new Jackson2JsonDecoder(mapper));
        defaults.jackson2JsonEncoder(new Jackson2JsonEncoder(mapper));
    }
}
