package com.flyfish.framework.handler;

import com.flyfish.framework.bean.BaseResponse;
import com.flyfish.framework.bean.Result;
import com.flyfish.framework.controller.SafeController;
import com.flyfish.framework.exception.BaseException;
import com.flyfish.framework.exception.auth.ClientTokenException;
import com.flyfish.framework.exception.auth.UserInvalidException;
import com.flyfish.framework.exception.auth.UserTokenException;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.context.support.DefaultMessageSourceResolvable;
import org.springframework.dao.DuplicateKeyException;
import org.springframework.http.HttpStatus;
import org.springframework.web.bind.annotation.ExceptionHandler;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.bind.annotation.ResponseStatus;
import org.springframework.web.bind.annotation.RestControllerAdvice;
import org.springframework.web.bind.support.WebExchangeBindException;
import reactor.core.publisher.Mono;

import java.util.stream.Collectors;


/**
 * Created by wangyu on 2017/9/8.
 * 只允许implement了SafeController的class
 */
@RestControllerAdvice(assignableTypes = SafeController.class)
@ResponseBody
public class GlobalExceptionHandler {
    private Logger logger = LoggerFactory.getLogger(GlobalExceptionHandler.class);

    @ExceptionHandler(ClientTokenException.class)
    public BaseResponse clientTokenExceptionHandler(ClientTokenException ex) {
        logger.error(ex.getMessage(), ex);
        return new BaseResponse(ex.getStatus(), ex.getMessage());
    }

    @ExceptionHandler(UserTokenException.class)
    public BaseResponse userTokenExceptionHandler(UserTokenException ex) {
        logger.error(ex.getMessage(), ex);
        return new BaseResponse(ex.getStatus(), ex.getMessage());
    }

    @ExceptionHandler(UserInvalidException.class)
    @ResponseStatus(HttpStatus.UNAUTHORIZED)
    public Mono<Result<Void>> userInvalidExceptionHandler(UserInvalidException ex) {
        logger.error(ex.getMessage(), ex);
        return Mono.just(Result.error(ex.getMessage()));
    }

    @ExceptionHandler(BaseException.class)
    public Mono<Result<Void>> baseExceptionHandler(BaseException ex) {
        logger.error(ex.getMessage(), ex);
        return Mono.just(Result.error(ex.getMessage()));
    }

    @ExceptionHandler(DuplicateKeyException.class)
    public Mono<Result<Void>> duplicateHandler(DuplicateKeyException ex) {
        logger.error(ex.getMessage(), ex);
        if (StringUtils.isNotBlank(ex.getMessage())) {
            if (ex.getMessage().contains("username")) {
                return Mono.just(Result.error("用户名已经被注册，请换一个哟！"));
            }
            if (ex.getMessage().contains("openId")) {
                return Mono.just(Result.error("本设备已经注册过帐号了，请更换设备再试哦！"));
            }
        }
        return Mono.just(Result.error("系统中存在重复的值了，请换一个哟"));
    }

    @ExceptionHandler(Exception.class)
    public Mono<Result<Void>> otherExceptionHandler(Exception ex) {
        logger.error(ex.getMessage(), ex);
        return Mono.just(Result.error(ex.getMessage()));
    }

    /**
     * 校验参数异常
     */
    @ExceptionHandler(WebExchangeBindException.class)
    public Mono<Result<Void>> validException(WebExchangeBindException e) {
        return Mono.just(Result.error(e.getBindingResult().getFieldErrors().stream().map(DefaultMessageSourceResolvable::getDefaultMessage)
                .collect(Collectors.joining("，"))));
    }
}
